// Copyright (C) 2021  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package blib

//go:generate ./gen-rangemap _RangedMap_str string zrangemap_str.go

import (
	"reflect"
	"testing"
)

func TestRangedMap(t *testing.T) {
	type RangedMap      = _RangedMap_str
	type RangedMapEntry = _RangedMap_strEntry

	type testEntry struct {
		A   *RangedMap
		B   RangedMapEntry
		Set *RangedMap // A.SetRange(B.keycov, B.value)
		Del *RangedMap // A.DelRange(B.keycov)
		Has bool       // A.HasRange(B.keycov)
	}
	E := func(A *RangedMap, B RangedMapEntry, S, D *RangedMap, H bool) testEntry {
		return testEntry{A, B, S, D, H}
	}

	// M is shorthand to create RangedMap, e.g. M(1,2,a, 3,4,b) will return {[1,2):a [3,4):b}.
	M := func(argv ...interface{}) *RangedMap {
		l := len(argv)
		if l % 3 != 0 {
			panic("non 3x number of arguments")
		}
		// askey converts arg to Key
		asKey := func(arg interface{}) Key {
			// go through reflect to accept all int, int64, Key, ...
			return Key(reflect.ValueOf(arg).Int())
		}
		M := &RangedMap{}
		for i := 0; i < l/3; i++ {
			// construct .entryv directly, not via SetRange
			lo := asKey(argv[3*i])
			hi := asKey(argv[3*i+1])
			v  := argv[3*i+2].(string)
			hi_ := hi
			if hi_ != oo {
				hi_--
			}
			M.entryv = append(M.entryv, RangedMapEntry{
					v,
					KeyRange{lo, hi_},
			})
		}
		M.verify()
		return M
	}

	// K creates RangedMapEntry{v, [lo,hi)}
	K := func(lo,hi Key, v string) RangedMapEntry {
		hi_ := hi
		if hi_ != noo {
			hi_--
		}
		return RangedMapEntry{v, KeyRange{lo, hi_}}
	}

	// y, n alias true/false
	const y, n = true, false

	// a, b, c, d ... are shorthands for corresponding strings
	const a, b, c, d, x = "a", "b", "c", "d", "x"


	testv := []testEntry{
		E(
			M(1,2,a, 2,3,b),	// A
			K(1,3,x),		// B
				M(1,3,x),		// Set
				M(),			// Del
				y),			// Has

		// coalesce (same value, no overlap)
		E(
			M(1,2,a, 4,5,a),	// A
			K(2,4,a),		// B
				M(1,5,a),		// Set
				M(1,2,a, 4,5,a),	// Del
				n),			// Has

		// coalesce (same value, overlap)
		E(
			M(1,4,a, 5,8,a),	// A
			K(2,6,a),		// B
				M(1,8,a),		// Set
				M(1,2,a, 6,8,a),	// Del
				n),			// Has

		// - shrink left/right (value !same) + new entry
		E(
			M(1,4,a),		// A
			K(2,6,x),		// B
				M(1,2,a, 2,6,x),	// Set
				M(1,2,a),		// Del
				n),			// Has
		E(
			M(5,8,b),		// A
			K(2,6,x),		// B
				M(2,6,x, 6,8,b),	// Set
				M(       6,8,b),	// Del
				n),			// Has
		E(
			M(1,4,a, 5,8,b),	// A
			K(2,6,x),		// B
				M(1,2,a, 2,6,x, 6,8,b),	// Set
				M(1,2,a,        6,8,b), // Del
				n),                     // Has



	}

	for _, tt := range testv {
		A := tt.A
		B := tt.B
		r := B.KeyRange
		v := B.Value

		assertMHasRange(t, A, r, tt.Has)
		Aset := A.Clone()
		Adel := A.Clone()
		Aset.SetRange(r, v)
		Adel.DelRange(r)

		if !Aset.Equal(tt.Set) {
			t.Errorf("SetRange:\n  A:   %s\n  e:   %s\n  ->·: %s\n  ok·: %s\n", A, B, Aset, tt.Set)

		}

		if !Adel.Equal(tt.Del) {
			t.Errorf("DelRange:\n  A:   %s\n  r:   %s\n  ->·: %s\n  ok·: %s\n", A, r, Adel, tt.Del)

		}

		assertMHasRange(t, Aset, r, true)
		assertMHasRange(t, Adel, r, false)

		verifyGet(t, A)
		verifyGet(t, Aset)
		verifyGet(t, Adel)
	}
}

// assertMHasRange asserts that RangeMap M.HasRange(r) == hasOK.
func assertMHasRange(t *testing.T, M *_RangedMap_str, r KeyRange, hasOK bool) {
	t.Helper()
	has := M.HasRange(r)
	if !(has == hasOK) {
		t.Errorf("HasRange:\n  M:   %s\n  r:   %s\n  ->·: %t\n  ok·: %t\n", M, r, has, hasOK)
	}
}

// verifyGet verifies RangedMap.Get .
func verifyGet(t *testing.T, M *_RangedMap_str) {
	type RangedMap      = _RangedMap_str
	type RangedMapEntry = _RangedMap_strEntry

	t.Helper()

	var Mranges []RangedMapEntry
	Mranges = append(Mranges, M.AllRanges()...) // copy just in case it changes on Get

	for _, e := range Mranges {
		for k := e.Lo; k <= e.Hi_; k++ {
			v, ok := M.Get_(k)
			if !(v == e.Value && ok) {
				t.Errorf("%s\tGet(%s):\nhave: %q, %t\nwant: %q, true",
					M, KStr(k), v, ok, e.Value)
			}
		}
	}

	// verify no-data
	// NA = [-∞,∞) \ M
	na := RangedKeySet{}
	na.AddRange(KeyRange{-10,+10}) // models -∞,∞
	for _, e := range Mranges {
		na.DelRange(e.KeyRange)
	}

	for _, r := range na.AllRanges() {
		for k := r.Lo; k <= r.Hi_; k++ {
			v, ok := M.Get_(k)
			if !(v == "" && !ok) {
				t.Errorf("%s\tGet(%s):\nhave: %q, %t\nwant: %q, false",
					M, KStr(k), v, ok, "")
			}
		}
	}
}
